﻿local PassLoot = LibStub("AceAddon-3.0"):GetAddon("PassLoot");
local L = LibStub("AceLocale-3.0"):GetLocale("PassLoot");

--[=[  Frame layout:
PassLoot.MainFrame = {
  ["Settings"] = Button,
  ["Done"] = Button,
  ["TabMenu"] = {
    ["Tab1"] = Tab,
    ["Tab2"] = Tab,
    ["Rules"] = {
      ["List"] = {  -- Has a background
        ["ScrollFrame"] = FauxScrollFrame,
        ["ScrollLine1"] = {
          ["Highlight"] = Highlight Texture,
          ["Text"] = FontString,
          ["Pass"] = CheckButton {
            ["Title"] = FontString,
          },
          ["Greed"] = CheckButton {
            ["Title"] = FontString,
          },
          ["Need"] = CheckButton {
            ["Title"] = FontString,
          },
        },
        ["ScrollLine6"],
        ["Add"] = Button,
        ["Remove"] = Button,
        ["Up"] = Button,
        ["Down"] = Button,
      },
      ["Settings"] = {  -- Has a background
        ["Desc"] = EditBox {
          ["Title"] = FontString,
        },
        ["AvailableFilters"] = {  -- Has a background
          ["Title"] = FontString,
          ["ScrollFrame"] = FauxScrollFrame,
          ["ScrollLine1"] = {
            ["Highlight"] = Highlight Texture,
            ["Text"] = FontString,
          },
          ["ScrollLine8"],
        },
        ["ActiveFilters"] = {  -- Has a background
          ["Title"] = FontString,
          ["ScrollFrame"] = FauxScrollFrame,
          ["ScrollLine1"] = {
            ["Highlight"] = Highlight Texture,
            ["Text"] = FontString,
          },
          ["ScrollLine8"],
        },
        ["Add"] = Button,
        ["Remove"] = Button,
        -- We insert rule widgets here, PassLoot.PluginInfo[ModuleName].RuleWidgets is a table of widgets per module.
      },
    },
    ["Modules"] = {  -- Has a background
      ["ScrollFrame"] = UIPanelScrollFrameTemplate {
        ["ScrollChild"] = Child Frame of variable size, has a background. {
          PassLoot.PluginInfo[ModuleName].ProfileHeader = {  -- Has a background
            ["CheckBox"] = CheckButton {
              ["Text"] = FontString,
            },
            ["Settings"] = {
              -- We insert profile widgets here, PassLoot.PluginInfo[ModuleName].ProfileWidgets is a table of them.
            },
          },
        },
      },
    },
  },
};
]=]

function PassLoot:ShowTooltip(...)
  if ( select("#", ...) == 0 ) then
    return;
  end
  GameTooltip:SetOwner(PassLoot_MainFrame, "ANCHOR_TOPLEFT");
  GameTooltip:SetText(PassLoot.FontWhite..select(1, ...));
  for i = 2, select("#", ...) do
    GameTooltip:AddLine(PassLoot.FontGold..select(i, ...));
  end
  GameTooltip:Show();
end

--Function to copy tables, since passing tables is always by reference.
function PassLoot:CopyTable(OldDB)
  if ( not OldDB or type(OldDB) ~= "table" ) then
    return OldDB;
  end
  local NewDB;
  NewDB = {};
  for Key, Value in pairs(OldDB) do
    if ( type(Value) ~= "table" ) then
      NewDB[Key] = Value;
    else
      NewDB[Key] = self:CopyTable(Value);
    end
  end
  return NewDB;
end

-- I am going to use this function to scroll text boxes to the left instead of SetCursorPosition()
-- SetCursorPosition(0) requires I ClearFocus(), which will create a loop that I don't really like.
function PassLoot:ScrollLeft(Frame, Elapsed)
  Frame:HighlightText(0,1);
  Frame:Insert(" "..strsub(Frame:GetText(),1,1));
  Frame:HighlightText(0,1);
  Frame:Insert("");
  Frame:SetScript("OnUpdate", nil);
end

function PassLoot:DisplayCurrentRule()
  if ( not self.CurrentRule ) then
    self.CurrentRule = 0;
  end
  self.CurrentOptionFilter = { nil, 0 }; -- Frame, line #
  self:DisplayCurrentOptionFilter();
  if ( self.CurrentRule > 0 ) then
    self.MainFrame.TabMenu.Rules.Settings.Desc:Show();
    self.MainFrame.TabMenu.Rules.Settings.AvailableFilters:Show();
    self.MainFrame.TabMenu.Rules.Settings.ActiveFilters:Show();
    self:Rules_AvailableFilters_OnScroll();
    self:Rules_ActiveFilters_OnScroll();
    self.MainFrame.TabMenu.Rules.Settings.Desc:SetText(self.db.profile.Rules[self.CurrentRule].Desc);
    self.MainFrame.TabMenu.Rules.Settings.Desc:SetScript("OnUpdate", function(...) self:ScrollLeft(...) end);
  else
    self.MainFrame.TabMenu.Rules.Settings.Desc:Hide();
    self.MainFrame.TabMenu.Rules.Settings.AvailableFilters:Hide();
    self.MainFrame.TabMenu.Rules.Settings.ActiveFilters:Hide();
  end
end

-- Show the widget that is selected
function PassLoot:DisplayCurrentOptionFilter()
  local Widget;
  if ( self.OldOptionFilter ) then
    self.OldOptionFilter:Hide();
  end
  if ( self.CurrentOptionFilter[1] == "Available" ) then
    self.MainFrame.TabMenu.Rules.Settings.Add:Show();
    self.MainFrame.TabMenu.Rules.Settings.Remove:Hide();
  elseif ( self.CurrentOptionFilter[1] == "Active" ) then
    local WidgetKey, Offset = self:GetWidgetFromLineNum(self.CurrentOptionFilter[2]);
    self.MainFrame.TabMenu.Rules.Settings.Add:Hide();
    self.MainFrame.TabMenu.Rules.Settings.Remove:Show();
    if ( Offset ) then
      Widget = self.RuleWidgets[WidgetKey];
      Widget:DisplayWidget(Offset);
      Widget:Show();
    end
  else
    self.MainFrame.TabMenu.Rules.Settings.Add:Hide();
    self.MainFrame.TabMenu.Rules.Settings.Remove:Hide();
  end
  self.OldOptionFilter = Widget;
end

function PassLoot:SetLootMethod(LineNum, Method)
  local Frame = self.MainFrame.TabMenu.Rules.List;
  local RuleNum = LineNum + FauxScrollFrame_GetOffset(Frame.ScrollFrame);
  local NewMethod = "disabled";
  Frame["ScrollLine"..LineNum].Pass:SetChecked(false);
  Frame["ScrollLine"..LineNum].Greed:SetChecked(false);
  Frame["ScrollLine"..LineNum].Need:SetChecked(false);
  if ( Method == "Pass" and self.db.profile.Rules[RuleNum].Loot ~= "pass" ) then
    NewMethod = "pass";
    Frame["ScrollLine"..LineNum].Pass:SetChecked(true);
  end
  if ( Method == "Greed" and self.db.profile.Rules[RuleNum].Loot ~= "greed" ) then
    NewMethod = "greed";
    Frame["ScrollLine"..LineNum].Greed:SetChecked(true);
  end
  if ( Method == "Need" and self.db.profile.Rules[RuleNum].Loot ~= "need" ) then
    NewMethod = "need";
    Frame["ScrollLine"..LineNum].Need:SetChecked(true);
  end
  self.db.profile.Rules[RuleNum].Loot = NewMethod;
  -- if ( Method == "Pass" ) then
    -- if ( self.db.profile.Rules[RuleNum].Loot == "pass" ) then
      -- self.db.profile.Rules[RuleNum].Loot = "disabled";
    -- else
      -- self.db.profile.Rules[RuleNum].Loot = "pass";
    -- end
    -- Frame["ScrollLine"..LineNum].Greed:SetChecked(false);
    -- Frame["ScrollLine"..LineNum].Need:SetChecked(false);
  -- elseif ( Method == "Greed" ) then
    -- if ( self.db.profile.Rules[RuleNum].Loot == "greed" ) then
      -- self.db.profile.Rules[RuleNum].Loot = "disabled";
    -- else
      -- self.db.profile.Rules[RuleNum].Loot = "greed";
    -- end
    -- Frame["ScrollLine"..LineNum].Pass:SetChecked(false);
    -- Frame["ScrollLine"..LineNum].Need:SetChecked(false);
  -- elseif ( Method == "Need" ) then
    -- if ( self.db.profile.Rules[RuleNum].Loot == "need" ) then
      -- self.db.profile.Rules[RuleNum].Loot = "disabled";
    -- else
      -- self.db.profile.Rules[RuleNum].Loot = "need";
    -- end
    -- Frame["ScrollLine"..LineNum].Pass:SetChecked(false);
    -- Frame["ScrollLine"..LineNum].Greed:SetChecked(false);
  -- end
end

function PassLoot:SetCurrentRule(LineNum)
  local Counter;
  local Frame = self.MainFrame.TabMenu.Rules.List;
  self.CurrentRule = LineNum + FauxScrollFrame_GetOffset(Frame.ScrollFrame);
  for Counter = 1, self.NumRuleListLines do
    if ( Counter == LineNum ) then
      Frame["ScrollLine"..Counter].Highlight:Show();
    else
      Frame["ScrollLine"..Counter].Highlight:Hide();
    end
  end
  self:DisplayCurrentRule();
end

function PassLoot:SetCurrentOptionFilter(FilterList, LineNum, Button)
  local Counter;
  self.CurrentOptionFilter = {
    FilterList,
    LineNum + FauxScrollFrame_GetOffset(self.MainFrame.TabMenu.Rules.Settings[FilterList.."Filters"].ScrollFrame),
  };
  if ( Button == "RightButton" and IsShiftKeyDown() and FilterList == "Active" ) then
    self:RemoveFilter();
  else
    for Counter = 1, self.NumFilterLines do
      if ( LineNum == Counter ) then
        if ( FilterList == "Available" ) then
          self.MainFrame.TabMenu.Rules.Settings.AvailableFilters["ScrollLine"..Counter].Highlight:Show();
          self.MainFrame.TabMenu.Rules.Settings.ActiveFilters["ScrollLine"..Counter].Highlight:Hide();
        else
          self.MainFrame.TabMenu.Rules.Settings.AvailableFilters["ScrollLine"..Counter].Highlight:Hide();
          self.MainFrame.TabMenu.Rules.Settings.ActiveFilters["ScrollLine"..Counter].Highlight:Show();
        end
      else
        self.MainFrame.TabMenu.Rules.Settings.AvailableFilters["ScrollLine"..Counter].Highlight:Hide();
        self.MainFrame.TabMenu.Rules.Settings.ActiveFilters["ScrollLine"..Counter].Highlight:Hide();
      end
    end
  end
  self:DisplayCurrentOptionFilter();
end

function PassLoot:RemoveFilter()
  if ( self.CurrentRule > 0 and self.CurrentOptionFilter[1] == "Active" and self.CurrentOptionFilter[2] > 0 ) then
    local WidgetKey, Offset = self:GetWidgetFromLineNum(self.CurrentOptionFilter[2]);
    if ( Offset ) then
      self.RuleWidgets[WidgetKey]:RemoveFilter(Offset);
    else
      for Index = self.RuleWidgets[WidgetKey]:GetNumFilters(), 1, -1 do
        self.RuleWidgets[WidgetKey]:RemoveFilter(Index);
      end
    end
    for Key, Value in pairs(self.RuleWidgets) do
      Value:Hide();
    end
    self.CurrentOptionFilter = { nil, 0 }; -- Frame, line #
    -- self:DisplayCurrentOptionFilter();
    self:Rules_AvailableFilters_OnScroll();
    self:Rules_ActiveFilters_OnScroll();
  end
end

function PassLoot:Rules_RuleList_OnScroll()
  local Frame = self.MainFrame.TabMenu.Rules.List;
  local Line, LineNum;
  local NumRules = #self.db.profile.Rules;
  FauxScrollFrame_Update(Frame.ScrollFrame, NumRules, self.NumRuleListLines, self.RuleListLineHeight);
  for Line=1, self.NumRuleListLines do
    LineNum = Line + FauxScrollFrame_GetOffset(Frame.ScrollFrame);
    if ( LineNum <= NumRules ) then
      Frame["ScrollLine"..Line].Text:SetText(self.db.profile.Rules[LineNum].Desc);
      Frame["ScrollLine"..Line].Pass:SetChecked(false);
      Frame["ScrollLine"..Line].Greed:SetChecked(false);
      Frame["ScrollLine"..Line].Need:SetChecked(false);
      if ( self.db.profile.Rules[LineNum].Loot == "pass" ) then
        Frame["ScrollLine"..Line].Pass:SetChecked(true);
      elseif ( self.db.profile.Rules[LineNum].Loot == "greed" ) then
        Frame["ScrollLine"..Line].Greed:SetChecked(true);
      elseif ( self.db.profile.Rules[LineNum].Loot == "need" ) then
        Frame["ScrollLine"..Line].Need:SetChecked(true);
      end
      Frame["ScrollLine"..Line]:Show();
      if ( LineNum == self.CurrentRule ) then
        Frame["ScrollLine"..Line].Highlight:Show();
      else
        Frame["ScrollLine"..Line].Highlight:Hide();
      end
    else
      Frame["ScrollLine"..Line].Highlight:Hide();
      Frame["ScrollLine"..Line]:Hide();
    end
  end
end

function PassLoot:Rules_AvailableFilters_OnScroll()
  local Frame = self.MainFrame.TabMenu.Rules.Settings.AvailableFilters;
  local Line, LineNum;
  local NumOptions = #self.RuleWidgets;
  FauxScrollFrame_Update(Frame.ScrollFrame, NumOptions, self.NumFilterLines, self.FilterLineHeight);
  for Line=1, self.NumFilterLines do
    LineNum = Line + FauxScrollFrame_GetOffset(Frame.ScrollFrame);
    if ( LineNum <= NumOptions ) then
      Frame["ScrollLine"..Line].Text:SetText(self.RuleWidgets[LineNum].Info[1] or "");
      Frame["ScrollLine"..Line]:Show();
      if ( self.CurrentOptionFilter[1] == "Available" and self.CurrentOptionFilter[2] == LineNum ) then
        Frame["ScrollLine"..Line].Highlight:Show();
      else
        Frame["ScrollLine"..Line].Highlight:Hide();
      end
    else
      Frame["ScrollLine"..Line].Highlight:Hide();
      Frame["ScrollLine"..Line]:Hide();
    end
  end
end

function PassLoot:Rules_ActiveFilters_OnScroll()
  if ( self.CurrentRule < 1 ) then
    return;
  end
  local Frame = self.MainFrame.TabMenu.Rules.Settings.ActiveFilters;
  local Line, LineNum;
  local NumLines = 0;
  local WidgetKey, Offset, Text, NumFilters;
  for WidgetKey, WidgetValue in ipairs(self.RuleWidgets) do
    NumFilters = (WidgetValue:GetNumFilters() or 0);
    if ( NumFilters > 0 ) then
      NumLines = NumLines + NumFilters + 1;
    end
  end
  FauxScrollFrame_Update(Frame.ScrollFrame, NumLines, self.NumFilterLines, self.FilterLineHeight);
  for Line=1, self.NumFilterLines do
    LineNum = Line + FauxScrollFrame_GetOffset(Frame.ScrollFrame);
    if ( LineNum <= NumLines ) then
      WidgetKey, Offset = self:GetWidgetFromLineNum(LineNum);
      if ( Offset ) then
        Text = self.RuleWidgets[WidgetKey]:GetFilterText(Offset) or "Value Error";
      else
        Text = PassLoot.FontGold..(self.RuleWidgets[WidgetKey].Info[1] or "Name Error");
      end
      Frame["ScrollLine"..Line].Text:SetText(Text);
      Frame["ScrollLine"..Line]:Show();
      if ( self.CurrentOptionFilter[1] == "Active" and self.CurrentOptionFilter[2] == LineNum ) then
        Frame["ScrollLine"..Line].Highlight:Show();
      else
        Frame["ScrollLine"..Line].Highlight:Hide();
      end
    else
      Frame["ScrollLine"..Line].Highlight:Hide();
      Frame["ScrollLine"..Line]:Hide();
    end
  end
end

function PassLoot:GetWidgetFromLineNum(Offset)
  local VariableList, NumFilters;
  for WidgetKey, WidgetValue in ipairs(self.RuleWidgets) do
    NumFilters = WidgetValue:GetNumFilters() or 0;
    if ( NumFilters > 0 ) then
      Offset = Offset - 1;
      if ( Offset == 0 ) then
        return WidgetKey, nil;
      end
    end
    if ( Offset <= NumFilters ) then
      return WidgetKey, Offset;
    else
      Offset = Offset - NumFilters;
    end
  end
end

function PassLoot:TabButtonClick(Frame, TabbedMenuContainer)
  if ( select(4, GetBuildInfo()) < 30000 ) then
    PanelTemplates_Tab_OnClick(TabbedMenuContainer);
  else
    PanelTemplates_Tab_OnClick(Frame, TabbedMenuContainer);
  end
  self.MainFrame.TabMenu.Rules:Hide();
  self.MainFrame.TabMenu.Modules:Hide();
  if ( TabbedMenuContainer.selectedTab == 1 ) then
    self.MainFrame.TabMenu.Rules:Show();
  elseif ( TabbedMenuContainer.selectedTab == 2 ) then
    self.MainFrame.TabMenu.Modules:Show();
  end
end

--[=[ ##########################
           START OF NEW CODE
      ##########################
]=]

function PassLoot:DisplayMainFrameAtAnchor()
  self.MainFrame:ClearAllPoints();
  if ( self.db.profile.WindowPos.X and self.db.profile.WindowPos.Y ) then
    local Scale = self.MainFrame:GetEffectiveScale();
    self.MainFrame:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", self.db.profile.WindowPos.X / Scale, self.db.profile.WindowPos.Y / Scale);
  else
    self.MainFrame:SetPoint("CENTER");
  end
end

-- Tooltip used for scanning (we pass the frame to modules, and they can scan)
function PassLoot:Create_PassLootTooltip()
  local Frame = CreateFrame("GameTooltip", "PassLootTooltip", UIParent, "GameTooltipTemplate");
  Frame:Hide();
  Frame:SetOwner(UIParent, "ANCHOR_NONE");
  return Frame;
end

function PassLoot:Create_MainFrame()
  local Frame = CreateFrame("Frame", "PassLoot_MainFrame", UIParent);
  Frame:SetWidth(480);
  Frame:SetHeight(505);
  Frame:SetBackdrop({
    ["bgFile"] = "Interface\\DialogFrame\\UI-DialogBox-Background",
    ["edgeFile"] = "Interface\\DialogFrame\\UI-DialogBox-Border",
    ["tile"] = true,
    ["insets"] = {
      ["top"] = 12,
      ["bottom"] = 11,
      ["left"] = 11,
      ["right"] = 12,
    },
    ["tileSize"] = 32,
    ["edgeSize"] = 32,
  });
  Frame:EnableMouse(true);
  Frame:EnableKeyboard(true);
  Frame:SetMovable(true);
  Frame:SetClampedToScreen(true);
  Frame:SetClampRectInsets(300, -300, -350, 350);
  Frame:SetFrameStrata("DIALOG");
  Frame:SetScript("OnMouseDown", function() Frame:StartMoving() end);
  Frame:SetScript("OnMouseUp", function()
    Frame:StopMovingOrSizing()
    local s = Frame:GetEffectiveScale()
    self.db.profile.WindowPos.X = Frame:GetLeft() * s
    self.db.profile.WindowPos.Y = Frame:GetTop() * s
  end);
  Frame:Hide();

  Frame.HeaderTexture = Frame:CreateTexture(nil, "ARTWORK");
  Frame.HeaderTexture:SetTexture("Interface\\DialogFrame\\UI-DialogBox-Header");
  Frame.HeaderTexture:SetWidth(320);
  Frame.HeaderTexture:SetHeight(64);
  Frame.HeaderTexture:SetPoint("TOP", Frame, "TOP", 0, 12);

  Frame.HeaderText = Frame:CreateFontString(nil, "ARTWORK", "GameFontNormal");
  Frame.HeaderText:SetPoint("TOP", Frame.HeaderTexture, "TOP", 0, -14);
  Frame.HeaderText:SetText(L["PassLoot"]);
  
  Frame.TabMenu = self:Create_TabMenu();
  Frame.TabMenu:SetParent(Frame);
  Frame.TabMenu:SetPoint("TOP", Frame, "TOP", 0, -47);
  -- Frame.TabMenu.Rules.List:SetScript("OnShow", function() self:Rules_RuleList_OnScroll() end);
  -- Frame.TabMenu.Rules.Settings:SetScript("OnShow", function() self:DisplayCurrentRule() end);
  
  Frame.Options = CreateFrame("Button", nil, Frame, "UIPanelButtonTemplate");
  Frame.Options:SetPoint("BOTTOMLEFT", Frame, "BOTTOMLEFT", 15, 15);
  Frame.Options:SetWidth(90); -- My other mods: 80
  Frame.Options:SetHeight(21); -- My other mods: 22
  Frame.Options:SetScript("OnEnter", function() self:ShowTooltip(L["Options"], L["Display the options window"]) end);
  Frame.Options:SetScript("OnLeave", function() GameTooltip:Hide() end);
  Frame.Options:SetScript("OnClick", function(frame, button) LibStub("AceConfigDialog-3.0"):Open(L["PassLoot"]) end);
  Frame.Options:SetText(L["Options"]);

  Frame.Done = CreateFrame("Button", nil, Frame, "UIPanelButtonTemplate");
  Frame.Done:SetPoint("BOTTOMRIGHT", Frame, "BOTTOMRIGHT", -15, 15);
  Frame.Done:SetWidth(90); -- My other mods: 80
  Frame.Done:SetHeight(21); -- My other mods: 22
  Frame.Done:SetScript("OnEnter", function() self:ShowTooltip(L["Done"], L["Close this window."]) end);
  Frame.Done:SetScript("OnLeave", function() GameTooltip:Hide() end);
  Frame.Done:SetScript("OnClick", function(frame, button) Frame:Hide() end);
  Frame.Done:SetText(L["Done"]);

  return Frame;
end

function PassLoot:Create_TabMenu()
  local Frame = CreateFrame("Frame", "PassLoot_TabbedMenuContainer");
  Frame:SetWidth(450);
  Frame:SetHeight(420);
  
  Frame.Tab1 = self:Create_TabButton(Frame, 1, L["Rules"], L["Configure rules."]);
  Frame.Tab1:SetParent(Frame);
  Frame.Tab1:SetPoint("TOPLEFT", Frame, "TOPLEFT", 5, 30);

  Frame.Tab2 = self:Create_TabButton(Frame, 2, L["Modules"], L["Configure modules."]);
  Frame.Tab2:SetParent(Frame);
  Frame.Tab2:SetPoint("LEFT", Frame.Tab1, "RIGHT", 2, 0);
  
  Frame.Rules = self:Create_RulesFrame();
  Frame.Rules:SetParent(Frame);
  Frame.Rules:SetPoint("TOP", Frame, "TOP");
  
  Frame.Modules = self:Create_ModulesFrame();
  Frame.Modules:SetParent(Frame);
  Frame.Modules:SetPoint("TOP", Frame, "TOP");
  Frame.Modules:Hide();
  
  return Frame;
end

function PassLoot:Create_TabButton(ParentFrame, ID, Name, Desc)
  local GlobalName = ParentFrame:GetName().."Tab"..ID;
  local Frame = CreateFrame("Button", GlobalName, nil, "TabButtonTemplate");
  Frame:SetID(ID);
  Frame:SetScript("OnClick", function(frame, button, down) self:TabButtonClick(frame, ParentFrame) end);
  Frame:SetScript("OnLeave", function() GameTooltip:Hide() end);
  Frame:SetScript("OnEnter", function(frame, motion) self:ShowTooltip(Name, Desc) end);
  Frame:SetText(Name);
  if ( select(4, GetBuildInfo()) >= 30000 ) then
    PanelTemplates_TabResize(Frame, 0);
  else
    self = Frame;
    PanelTemplates_TabResize(0);
  end
  return Frame;
end

function PassLoot:Create_RulesFrame()
  local Frame = CreateFrame("Frame");
  Frame:SetWidth(450);
  Frame:SetHeight(420);
  
  Frame.List = self:Create_RuleListFrame();
  Frame.List:SetParent(Frame);
  Frame.List:SetPoint("TOP", Frame, "TOP");

  Frame.Settings = self:Create_RuleSettingsFrame();
  Frame.Settings:SetParent(Frame);
  Frame.Settings:SetPoint("TOP", Frame.List, "BOTTOM");
  
  return Frame;
end

function PassLoot:Create_RuleListFrame()
  local Frame = CreateFrame("Frame");
  Frame:SetWidth(450);
  Frame:SetHeight(130);
  Frame:SetBackdrop({
    ["bgFile"] = "Interface\\Tooltips\\UI-Tooltip-Background",
    ["edgeFile"] = "Interface\\Tooltips\\UI-Tooltip-Border",
    ["tile"] = true,
    ["insets"] = {
      ["top"] = 5,
      ["bottom"] = 5,
      ["left"] = 5,
      ["right"] = 5,
    },
    ["tileSize"] = 16,
    ["edgeSize"] = 16,
  });
  Frame:SetBackdropBorderColor(0.4, 0.4, 0.4);
  Frame:SetBackdropColor(0.5, 0.5, 0.5);

  Frame.ScrollFrame = CreateFrame("ScrollFrame", "PassLoot_Rules_Scroll", Frame, "FauxScrollFrameTemplate");
  Frame.ScrollFrame:SetPoint("TOPLEFT", Frame, "TOPLEFT", 3, -8);
  Frame.ScrollFrame:SetWidth(415);
  Frame.ScrollFrame:SetHeight(96);
  if ( select(4, GetBuildInfo()) >= 30000 ) then
    Frame.ScrollFrame:SetScript("OnVerticalScroll", function(frame, offset)
      FauxScrollFrame_OnVerticalScroll(frame, offset, 16, function() self:Rules_RuleList_OnScroll() end)
    end);
  else
    Frame.ScrollFrame:SetScript("OnVerticalScroll", function()
      FauxScrollFrame_OnVerticalScroll(16, function() self:Rules_RuleList_OnScroll() end)
    end);
  end
  
  Frame.ScrollLine1 = self:Create_RuleListScrollLine();
  Frame.ScrollLine1:SetParent(Frame);
  Frame.ScrollLine1:SetPoint("TOPLEFT", Frame.ScrollFrame, "TOPLEFT", 8, 0);
  Frame.ScrollLine1.LineNum = 1;
  for Index = 2, 6 do
    Frame["ScrollLine"..Index] = self:Create_RuleListScrollLine();
    Frame["ScrollLine"..Index]:SetParent(Frame);
    Frame["ScrollLine"..Index]:SetPoint("TOPLEFT", Frame["ScrollLine"..(Index - 1)], "BOTTOMLEFT");
    Frame["ScrollLine"..Index].LineNum = Index;
  end
  
  Frame.Add = CreateFrame("Button", nil, Frame, "UIPanelButtonTemplate");
  Frame.Add:SetPoint("TOPLEFT", Frame.ScrollFrame, "BOTTOMLEFT", 30, 0);
  Frame.Add:SetWidth(90); -- My other mods: 80
  Frame.Add:SetHeight(21); -- My other mods: 22
  Frame.Add:SetScript("OnEnter", function() self:ShowTooltip(L["Add"], L["Add a new rule."]) end);
  Frame.Add:SetScript("OnLeave", function() GameTooltip:Hide() end);
  Frame.Add:SetScript("OnClick", function(frame, button)
    local TempDB = {}
    for Key, Value in ipairs(self.DefaultTemplate) do
      TempDB[Value[1]] = self:CopyTable(Value[2])
    end
    table.insert(self.db.profile.Rules, TempDB)
    self:Rules_RuleList_OnScroll()
  end);
  Frame.Add:SetText(L["Add"]);
  
  Frame.Remove = CreateFrame("Button", nil, Frame, "UIPanelButtonTemplate");
  Frame.Remove:SetPoint("TOPLEFT", Frame.Add, "TOPRIGHT", 10, 0);
  Frame.Remove:SetWidth(90); -- My other mods: 80
  Frame.Remove:SetHeight(21); -- My other mods: 22
  Frame.Remove:SetScript("OnEnter", function() self:ShowTooltip(L["Remove"], L["Remove selected rule."]) end);
  Frame.Remove:SetScript("OnLeave", function() GameTooltip:Hide() end);
  Frame.Remove:SetScript("OnClick", function(frame, button)
    if ( self.CurrentRule > 0 ) then
      table.remove(self.db.profile.Rules, self.CurrentRule)
      self.CurrentRule = 0
      self:Rules_RuleList_OnScroll()
      self:DisplayCurrentRule()
    end
  end);
  Frame.Remove:SetText(L["Remove"]);
  
  Frame.Up = CreateFrame("Button", nil, Frame, "UIPanelButtonTemplate");
  Frame.Up:SetPoint("TOPLEFT", Frame.Remove, "TOPRIGHT", 10, 0);
  Frame.Up:SetWidth(90); -- My other mods: 80
  Frame.Up:SetHeight(21); -- My other mods: 22
  Frame.Up:SetScript("OnEnter", function() self:ShowTooltip(L["Up"], L["Move selected rule up in priority."]) end);
  Frame.Up:SetScript("OnLeave", function() GameTooltip:Hide() end);
  Frame.Up:SetScript("OnClick", function(frame, button)
    if ( self.CurrentRule > 1 ) then
      local TempDB = self.db.profile.Rules[self.CurrentRule]
      self.db.profile.Rules[self.CurrentRule] = self.db.profile.Rules[self.CurrentRule - 1]
      self.db.profile.Rules[self.CurrentRule - 1] = TempDB
      self.CurrentRule = self.CurrentRule - 1
      self:Rules_RuleList_OnScroll()
      self:DisplayCurrentRule()
    end
  end);
  Frame.Up:SetText(L["Up"]);

  Frame.Down = CreateFrame("Button", nil, Frame, "UIPanelButtonTemplate");
  Frame.Down:SetPoint("TOPLEFT", Frame.Up, "TOPRIGHT", 10, 0);
  Frame.Down:SetWidth(90); -- My other mods: 80
  Frame.Down:SetHeight(21); -- My other mods: 22
  Frame.Down:SetScript("OnEnter", function() self:ShowTooltip(L["Down"], L["Move selected rule down in priority."]) end);
  Frame.Down:SetScript("OnLeave", function() GameTooltip:Hide() end);
  Frame.Down:SetScript("OnClick", function(frame, button)
    if ( self.CurrentRule > 0 and self.CurrentRule < #self.db.profile.Rules ) then
      local TempDB = self.db.profile.Rules[self.CurrentRule]
      self.db.profile.Rules[self.CurrentRule] = self.db.profile.Rules[self.CurrentRule + 1]
      self.db.profile.Rules[self.CurrentRule + 1] = TempDB
      self.CurrentRule = self.CurrentRule + 1
      self:Rules_RuleList_OnScroll()
      self:DisplayCurrentRule()
    end
  end);
  Frame.Down:SetText(L["Down"]);

  return Frame;
end

function PassLoot:Create_RuleListScrollLine()
  local Frame = CreateFrame("Button");
  Frame:SetWidth(413);
  Frame:SetHeight(16);
  Frame:SetScript("OnEnter", function(frame) self:ShowTooltip(L["Rule List"], L["Click to select and edit this rule."]) end);
  Frame:SetScript("OnLeave", function() GameTooltip:Hide() end);
  Frame:SetScript("OnClick", function(frame, button) self:SetCurrentRule(Frame.LineNum) end);
  
  Frame.Highlight = Frame:CreateTexture(nil, "BACKGROUND");
  Frame.Highlight:SetTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight");
  Frame.Highlight:SetAllPoints(Frame);
  Frame.Highlight:SetBlendMode("ADD");
  Frame.Highlight:Hide();

  Frame.Text = Frame:CreateFontString(nil, "BACKGROUND", "ChatFontNormal");
  Frame.Text:SetPoint("TOPLEFT", Frame, "TOPLEFT", 0, 0);
  Frame.Text:SetWidth(170);
  Frame.Text:SetHeight(16);
  Frame.Text:SetJustifyH("LEFT");
  
  Frame.Pass = self:Create_CheckBox();
  Frame.Pass:SetParent(Frame);
  Frame.Pass:SetPoint("TOPLEFT", Frame.Text, "TOPRIGHT");
  Frame.Pass:SetScript("OnClick", function(frame, button) self:SetLootMethod(Frame.LineNum, "Pass") end);
  Frame.Pass:SetScript("OnEnter", function() self:ShowTooltip(L["Pass"], L["Will pass on all loot matching this rule."]) end);
  Frame.Pass.Text:SetText(L["Pass"]);

  Frame.Greed = self:Create_CheckBox();
  Frame.Greed:SetParent(Frame);
  Frame.Greed:SetPoint("TOPLEFT", Frame.Pass, "TOPRIGHT", 60, 0);
  Frame.Greed:SetScript("OnClick", function(frame, button) self:SetLootMethod(Frame.LineNum, "Greed") end);
  Frame.Greed:SetScript("OnEnter", function() self:ShowTooltip(L["Greed"], L["Will roll greed on all loot matching this rule."]) end);
  Frame.Greed.Text:SetText(L["Greed"]);

  Frame.Need = self:Create_CheckBox();
  Frame.Need:SetParent(Frame);
  Frame.Need:SetPoint("TOPLEFT", Frame.Greed, "TOPRIGHT", 60, 0);
  Frame.Need:SetScript("OnClick", function(frame, button)self:SetLootMethod(Frame.LineNum, "Need") end);
  Frame.Need:SetScript("OnEnter", function() self:ShowTooltip(L["Need"], L["Will roll need on all loot matching this rule."]) end);
  Frame.Need.Text:SetText(L["Need"]);
  return Frame;
end

function PassLoot:Create_CheckBox()
  local Frame = CreateFrame("CheckButton");
  Frame:SetHeight(16);
  Frame:SetWidth(16);
  Frame:SetNormalTexture("Interface\\Buttons\\UI-CheckBox-Up");
  Frame:SetPushedTexture("Interface\\Buttons\\UI-CheckBox-Down");
  Frame:SetHighlightTexture("Interface\\Buttons\\UI-CheckBox-Highlight");
  Frame:SetCheckedTexture("Interface\\Buttons\\UI-CheckBox-Check");
  Frame:SetScript("OnLeave", function() GameTooltip:Hide() end);
  Frame:SetHitRectInsets(0, -35, 0, 0);
  Frame.Text = Frame:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall");
  Frame.Text:SetPoint("LEFT", Frame, "RIGHT", -2, 0);
  return Frame;
end

function PassLoot:Create_RuleSettingsFrame()
  local Frame = CreateFrame("Frame");
  Frame:SetWidth(450);
  Frame:SetHeight(290);
  Frame:SetBackdrop({
    ["bgFile"] = "Interface\\Tooltips\\UI-Tooltip-Background",
    ["edgeFile"] = "Interface\\Tooltips\\UI-Tooltip-Border",
    ["tile"] = true,
    ["insets"] = {
      ["top"] = 5,
      ["bottom"] = 5,
      ["left"] = 5,
      ["right"] = 5,
    },
    ["tileSize"] = 16,
    ["edgeSize"] = 16,
  });
  Frame:SetBackdropBorderColor(0.4, 0.4, 0.4);
  Frame:SetBackdropColor(0.5, 0.5, 0.5);
  
  Frame.Desc = self:Create_EditBox();
  Frame.Desc:SetParent(Frame);
  Frame.Desc:SetPoint("TOP", Frame, "TOP", 0, -15);
  Frame.Desc:SetWidth(160);
  Frame.Desc:SetHeight(26);
  Frame.Desc.Title:SetText(L["Description"]);
  Frame.Desc:SetScript("OnEnter", function(frame) self:ShowTooltip(L["Description"], L["Description_Desc"]) end);
  Frame.Desc:SetScript("OnEnterPressed", function(frame)
    if ( self.CurrentRule > 0 ) then
      self.db.profile.Rules[self.CurrentRule].Desc = frame:GetText();
    end
    frame:ClearFocus();
    self:Rules_RuleList_OnScroll();
  end);
  
  Frame.AvailableFilters = self:Create_RuleAvailableFiltersFrame();
  Frame.AvailableFilters:SetParent(Frame);
  Frame.AvailableFilters:SetPoint("TOPRIGHT", Frame.Desc, "BOTTOM", -30, -10);
  
  Frame.ActiveFilters = self:Create_RuleActiveFiltersFrame();
  Frame.ActiveFilters:SetParent(Frame);
  Frame.ActiveFilters:SetPoint("TOPLEFT", Frame.Desc, "BOTTOM", -30, -10);

  Frame.Add = CreateFrame("Button", nil, Frame, "UIPanelButtonTemplate");
  Frame.Add:SetPoint("TOPLEFT", Frame.AvailableFilters, "BOTTOMLEFT", 0, -5);
  Frame.Add:SetWidth(90); -- My other mods: 80
  Frame.Add:SetHeight(21); -- My other mods: 22
  Frame.Add:SetScript("OnEnter", function() self:ShowTooltip(L["Add"], L["Add this filter."]) end);
  Frame.Add:SetScript("OnLeave", function() GameTooltip:Hide() end);
  Frame.Add:SetScript("OnClick", function(frame, button)
    if ( self.CurrentRule > 0 and self.CurrentOptionFilter[1] == "Available" and self.CurrentOptionFilter[2] > 0 ) then
      self.RuleWidgets[self.CurrentOptionFilter[2]]:AddNewFilter()
      self:Rules_ActiveFilters_OnScroll()
    end
  end);
  Frame.Add:SetText(L["Add"]);
  Frame.Add:Hide();
  
  Frame.Remove = CreateFrame("Button", nil, Frame, "UIPanelButtonTemplate");
  Frame.Remove:SetPoint("TOPRIGHT", Frame.ActiveFilters, "BOTTOMRIGHT", 0, -5);
  Frame.Remove:SetWidth(90); -- My other mods: 80
  Frame.Remove:SetHeight(21); -- My other mods: 22
  Frame.Remove:SetScript("OnEnter", function() self:ShowTooltip(L["Remove"], L["Remove this filter."]) end);
  Frame.Remove:SetScript("OnLeave", function() GameTooltip:Hide() end);
  Frame.Remove:SetScript("OnClick", function(frame, button)
    self:RemoveFilter()
  end);
  Frame.Remove:SetText(L["Remove"]);
  Frame.Remove:Hide();
  
  return Frame;
end

function PassLoot:Create_EditBox()
  local Frame = CreateFrame("EditBox");
  Frame:SetBackdrop({
    ["bgFile"] = "Interface\\Tooltips\\UI-Tooltip-Background",
    ["edgeFile"] = "Interface\\Tooltips\\UI-Tooltip-Border",
    ["tile"] = true,
    ["insets"] = {
      ["top"] = 5,
      ["bottom"] = 5,
      ["left"] = 5,
      ["right"] = 5,
    },
    ["tileSize"] = 32,
    ["edgeSize"] = 16,
  });
  Frame:SetBackdropColor(0, 0, 0, 0.95);
  Frame:EnableMouse(true);
  Frame:SetMaxLetters(200);
  -- Frame:SetHistoryLines(0);
  Frame:SetAutoFocus(false);
  Frame:SetFontObject("ChatFontNormal");
  Frame:SetTextInsets(6, 6, 6, 6);
  Frame:SetScript("OnLeave", function() GameTooltip:Hide() end);
  Frame:SetScript("OnEscapePressed", function() Frame:ClearFocus() end);
  Frame:SetScript("OnEditFocusGained", function() Frame:HighlightText() end);
  Frame:SetScript("OnEditFocusLost", function()
    Frame:HighlightText(0, 0)
    self:DisplayCurrentRule()
  end);
  
  Frame.Title = Frame:CreateFontString(nil, "BACKGROUND", "GameFontNormalSmall");
  Frame.Title:SetPoint("BOTTOMLEFT", Frame, "TOPLEFT", 3, 0);
  
  return Frame;
end

function PassLoot:Create_RuleAvailableFiltersFrame()
  local Frame = CreateFrame("Frame");
  Frame:SetWidth(190);
  Frame:SetHeight(137);
  Frame:SetBackdrop({
    ["bgFile"] = "Interface\\DialogFrame\\UI-DialogBox-Background",
    ["edgeFile"] = "Interface\\Tooltips\\UI-Tooltip-Border",
    ["tile"] = true,
    ["insets"] = {
      ["top"] = 2,
      ["bottom"] = 2,
      ["left"] = 2,
      ["right"] = 2,
    },
    ["tileSize"] = 16,
    ["edgeSize"] = 16,
  });

  Frame.Title = Frame:CreateFontString(nil, "BACKGROUND", "GameFontNormalSmall");
  Frame.Title:SetPoint("BOTTOMLEFT", Frame, "TOPLEFT", 3, 0);
  Frame.Title:SetText(L["Available Filters"]);

  Frame.ScrollFrame = CreateFrame("ScrollFrame", "PassLoot_AvailableFilters_Scroll", Frame, "FauxScrollFrameTemplate");
  Frame.ScrollFrame:SetPoint("TOPLEFT", Frame, "TOPLEFT", 0, -5);
  Frame.ScrollFrame:SetWidth(162);
  Frame.ScrollFrame:SetHeight(128);
  if ( select(4, GetBuildInfo()) >= 30000 ) then
    Frame.ScrollFrame:SetScript("OnVerticalScroll", function(frame, offset)
      FauxScrollFrame_OnVerticalScroll(frame, offset, 16, function() self:Rules_AvailableFilters_OnScroll() end)
    end);
  else
    Frame.ScrollFrame:SetScript("OnVerticalScroll", function()
      FauxScrollFrame_OnVerticalScroll(16, function() self:Rules_AvailableFilters_OnScroll() end)
    end);
  end
  
  Frame.ScrollLine1 = self:Create_AvailableFiltersScrollLine();
  Frame.ScrollLine1:SetParent(Frame);
  Frame.ScrollLine1:SetPoint("TOPLEFT", Frame.ScrollFrame, "TOPLEFT", 8, 0);
  Frame.ScrollLine1.LineNum = 1;
  for Index = 2, 8 do
    Frame["ScrollLine"..Index] = self:Create_AvailableFiltersScrollLine();
    Frame["ScrollLine"..Index]:SetParent(Frame);
    Frame["ScrollLine"..Index]:SetPoint("TOPLEFT", Frame["ScrollLine"..(Index - 1)], "BOTTOMLEFT");
    Frame["ScrollLine"..Index].LineNum = Index;
  end

  return Frame;
end

function PassLoot:Create_AvailableFiltersScrollLine()
  local Frame = CreateFrame("Button");
  Frame:SetWidth(162);
  Frame:SetHeight(16);
  Frame:SetScript("OnEnter", function(frame) self:ShowTooltip(L["Available Filters"], L["Available Filters_Desc"]) end);
  Frame:SetScript("OnLeave", function() GameTooltip:Hide() end);
  Frame:SetScript("OnClick", function(frame, button) self:SetCurrentOptionFilter("Available", Frame.LineNum, button) end);

  Frame.Highlight = Frame:CreateTexture(nil, "BACKGROUND");
  Frame.Highlight:SetTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight");
  Frame.Highlight:SetAllPoints(Frame);
  Frame.Highlight:SetBlendMode("ADD");
  Frame.Highlight:Hide();

  Frame.Text = Frame:CreateFontString(nil, "BACKGROUND", "ChatFontNormal");
  Frame.Text:SetWidth(162);
  Frame.Text:SetHeight(16);
  Frame.Text:SetPoint("TOPLEFT", Frame, "TOPLEFT", 0, 0);
  Frame.Text:SetJustifyH("LEFT");
  return Frame;  
end

function PassLoot:Create_RuleActiveFiltersFrame()
  local Frame = CreateFrame("Frame");
  Frame:SetWidth(250);
  Frame:SetHeight(137);
  Frame:SetBackdrop({
    ["bgFile"] = "Interface\\DialogFrame\\UI-DialogBox-Background",
    ["edgeFile"] = "Interface\\Tooltips\\UI-Tooltip-Border",
    ["tile"] = true,
    ["insets"] = {
      ["top"] = 2,
      ["bottom"] = 2,
      ["left"] = 2,
      ["right"] = 2,
    },
    ["tileSize"] = 16,
    ["edgeSize"] = 16,
  });

  Frame.Title = Frame:CreateFontString(nil, "BACKGROUND", "GameFontNormalSmall");
  Frame.Title:SetPoint("BOTTOMLEFT", Frame, "TOPLEFT", 3, 0);
  Frame.Title:SetText(L["Active Filters"]);
  
  Frame.ScrollFrame = CreateFrame("ScrollFrame", "PassLoot_ActiveFilters_Scroll", Frame, "FauxScrollFrameTemplate");
  Frame.ScrollFrame:SetPoint("TOPLEFT", Frame, "TOPLEFT", 0, -5);
  Frame.ScrollFrame:SetWidth(222);
  Frame.ScrollFrame:SetHeight(128);
  if ( select(4, GetBuildInfo()) >= 30000 ) then
    Frame.ScrollFrame:SetScript("OnVerticalScroll", function(frame, offset)
      FauxScrollFrame_OnVerticalScroll(frame, offset, 16, function() self:Rules_ActiveFilters_OnScroll() end)
    end);
  else
    Frame.ScrollFrame:SetScript("OnVerticalScroll", function()
      FauxScrollFrame_OnVerticalScroll(16, function() self:Rules_ActiveFilters_OnScroll() end)
    end);
  end

  Frame.ScrollLine1 = self:Create_ActiveFiltersScrollLine();
  Frame.ScrollLine1:SetParent(Frame);
  Frame.ScrollLine1:SetPoint("TOPLEFT", Frame.ScrollFrame, "TOPLEFT", 8, 0);
  Frame.ScrollLine1.LineNum = 1;
  for Index = 2, 8 do
    Frame["ScrollLine"..Index] = self:Create_ActiveFiltersScrollLine();
    Frame["ScrollLine"..Index]:SetParent(Frame);
    Frame["ScrollLine"..Index]:SetPoint("TOPLEFT", Frame["ScrollLine"..(Index - 1)], "BOTTOMLEFT");
    Frame["ScrollLine"..Index].LineNum = Index;
  end

  return Frame;
end

function PassLoot:Create_ActiveFiltersScrollLine()
  local Frame = CreateFrame("Button");
  Frame:SetWidth(222);
  Frame:SetHeight(16);
  Frame:SetScript("OnEnter", function(frame) self:ShowTooltip(L["Active Filters"], L["Active Filters_Desc"]) end);
  Frame:SetScript("OnLeave", function() GameTooltip:Hide() end);
  Frame:SetScript("OnClick", function(frame, button) self:SetCurrentOptionFilter("Active", Frame.LineNum, button) end);
  Frame:RegisterForClicks("LeftButtonUp", "RightButtonUp");

  Frame.Highlight = Frame:CreateTexture(nil, "BACKGROUND");
  Frame.Highlight:SetTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight");
  Frame.Highlight:SetAllPoints(Frame);
  Frame.Highlight:SetBlendMode("ADD");
  Frame.Highlight:Hide();

  Frame.Text = Frame:CreateFontString(nil, "BACKGROUND", "ChatFontNormal");
  Frame.Text:SetWidth(222);
  Frame.Text:SetHeight(16);
  Frame.Text:SetPoint("TOPLEFT", Frame, "TOPLEFT", 0, 0);
  Frame.Text:SetJustifyH("LEFT");
  return Frame;  
end

function PassLoot:Create_ModulesFrame()
  local Frame = CreateFrame("Frame");
  Frame:SetWidth(450);
  Frame:SetHeight(420);
  Frame:SetBackdrop({
    ["bgFile"] = "Interface\\Tooltips\\UI-Tooltip-Background",
    ["edgeFile"] = "Interface\\Tooltips\\UI-Tooltip-Border",
    ["tile"] = true,
    ["insets"] = {
      ["top"] = 5,
      ["bottom"] = 5,
      ["left"] = 5,
      ["right"] = 5,
    },
    ["tileSize"] = 16,
    ["edgeSize"] = 16,
  });
  Frame:SetBackdropBorderColor(0.4, 0.4, 0.4);
  Frame:SetBackdropColor(0.5, 0.5, 0.5);

  Frame.ScrollFrame = CreateFrame("ScrollFrame", "PassLoot_Modules_ScrollFrame", Frame, "UIPanelScrollFrameTemplate"); -- Do I need to name it?
  Frame.ScrollFrame:SetPoint("TOPLEFT", Frame, "TOPLEFT", 5, -5);
  Frame.ScrollFrame:SetWidth(415);
  Frame.ScrollFrame:SetHeight(410);
  
  Frame.ScrollFrame.ScrollChild = CreateFrame("Frame", "PassLoot_Modules_ScrollFrame_Content");  -- Do I need to name it?
  Frame.ScrollFrame.ScrollChild:SetParent(Frame.ScrollFrame);
  Frame.ScrollFrame.ScrollChild:SetPoint("TOPLEFT", Frame.ScrollFrame, "TOPLEFT");
  Frame.ScrollFrame.ScrollChild:SetPoint("TOPRIGHT", Frame.ScrollFrame, "TOPRIGHT");
  Frame.ScrollFrame.ScrollChild:SetWidth(415);
  Frame.ScrollFrame.ScrollChild:SetHeight(600);
  Frame.ScrollFrame.ScrollChild:SetBackdrop({
    ["bgFile"] = "Interface\\DialogFrame\\UI-DialogBox-Background",
    ["edgeFile"] = "Interface\\Tooltips\\UI-Tooltip-Border",
    ["tile"] = true,
    ["insets"] = {
      ["top"] = 2,
      ["bottom"] = 2,
      ["left"] = 2,
      ["right"] = 2,
    },
    ["tileSize"] = 16,
    ["edgeSize"] = 16,
  });
  
  Frame.ScrollFrame:SetScrollChild(Frame.ScrollFrame.ScrollChild);
  return Frame;
end

--[=[
  <Frame name="OdlawMarkerFrames_Templates_TitledFrame" virtual="true">
    <Layers>
      <Layer level="BACKGROUND">
        <FontString name="$parentTitle" inherits="GameFontHighlight">
          <Anchors>
            <Anchor point="BOTTOMLEFT" relativePoint="TOPLEFT">
              <Offset> <AbsDimension x="9" y="0"/> </Offset>
            </Anchor>
          </Anchors>
        </FontString>
      </Layer>
    </Layers>
    <Backdrop bgFile="Interface\Tooltips\UI-Tooltip-Background" edgeFile="Interface\Tooltips\UI-Tooltip-Border" tile="true">
      <EdgeSize> <AbsValue val="16"/> </EdgeSize>
      <TileSize> <AbsValue val="16"/> </TileSize>
      <BackgroundInsets> <AbsInset left="5" right="5" top="5" bottom="5"/> </BackgroundInsets>
    </Backdrop>
    <Scripts>
      <OnLoad>
        self:SetBackdropBorderColor(0.4, 0.4, 0.4);
        self:SetBackdropColor(0.5, 0.5, 0.5);
      </OnLoad>
    </Scripts>
  </Frame>
]=]
